#include "declarations.h"

#if SIMULATOR
#include <stdio.h>
#else
#include <avr/interrupt.h>
#include <avr/sleep.h>
#endif

extern u8 *song_data;

u16 jiffies;
u16 song_pos;
u16	repeat_pos;
u8	repeats;
u8	time2event;
u8	time2control;
u8	clear_leds = 0;

#if SIMULATOR
volatile u8 control_signal = 1;
#else
#define control_signal GPIOR0
#endif

u8	sin_index;
u8	chan0_mod;
#define SIN_LEN 16
const s8 sin_table[SIN_LEN] = {5,14,21,24,24,21,14,5,-5,-14,-21,-24,-24,-21,-14,-5};
const s8 sin_table_offset[SIN_LEN] = {0, 3, 10, 19, 29, 38, 45, 48, 48, 45, 38, 29, 19, 10, 3, 0};

// menu variables
u8	volume = 192;
u8	demo = SIMULATOR;
u8	split_key = 0;
u8	trigger_val = SAMPLE_RATE / CONTROL_RATE / 2; // 2 triggers per control for LCD update
char *version_string = "R_Synth v3";

u8 read_song_byte(u16 pos);
u16 read_song_word(u16 pos);

// +==========================+
// |    Sound Functionality   |
// +==========================+

extern s8 wave_table1[WAVET_SIZE];
extern s8 wave_table2[WAVET_SIZE];

channel g_channels[NUM_CHANNELS];

// Functions

inline void env_reset(envelope *env) {
	env->state = env->init_state;
	if(env->initial < env->max)
		env->value = env->initial;
	else
		env->value = env->max;
	//env->time = 0;
}

inline void env_update(envelope *env) {
	s16 new_val;

	switch(env->state) {
	case ENV_ATTACK:
		new_val = (s16)env->value + env->attack;
		if(new_val > env->max) {
			env->value = env->max;
			env->state = ENV_DECAY;
			//env->time = 0;
		} else {
			//env->time++;
			env->value = (u8)new_val;
		}
		break;
	case ENV_DECAY:
		new_val = env->value - env->decay;
		if(new_val < env->hold) {
			env->state = ENV_HOLD;
			env->value = env->hold;
		} else {
			//env->time++;
			env->value = (u8)new_val;
		}
		break;
	case ENV_HOLD:
		env->value = env->hold;
		break;
	}
}

inline void env_update_all() {	// used only once hence inline
	u8 i, blend_env;
	static u8 lfo_index;
	u8 lfo_val = sin_table_offset[(lfo_index++ >> 4) & (SIN_LEN - 1)];

	for(i=0 ; i<NUM_CHANNELS; i++) {
		u8 vol, pan;
		s16 dif;
		channel *chan = &g_channels[i];

		env_update(&chan->env_vol);
		env_update(&chan->env_pan);
		vol = chan->env_vol.value * chan->vol >> 8;
		vol = vol * volume >> 8;
		pan = chan->env_pan.value;

		chan->inst_vol[0] = vol * (pan > 127 ? 255 : pan << 1) >> 8;
		chan->inst_vol[1] = vol * (pan < 128 ? 255 : (255 - pan) << 1) >> 8;

		env_update(&chan->osc_c.blend);

		blend_env = chan->lfo_scale * lfo_val;
		if(blend_env > chan->osc_c.blend.value)
			chan->osc_c.blend_val = 0;
		else
			chan->osc_c.blend_val = chan->osc_c.blend.value - blend_env;

		// mono slide
		dif = (s16)chan->osc_c.target_ds - chan->osc_c.ds;
		if(dif < 4 && dif > -4)
			chan->osc_c.ds += dif;
		else
			chan->osc_c.ds += dif>>2;
	}
#if DRUM_CHAN > 0
	env_update(&g_channels[DRUM_CHAN].osc_c.blend);
#endif
}

// Functions below are called at SAMPLE_RATE Hz

inline void update_osc(oscillator* osc) {
	osc->cur_sample[0] += osc->ds;
	osc->cur_sample[1] += osc->ds - osc->unison_shift;
#if UNISON_VOICES > 2
	osc->cur_sample[2] += osc->ds + osc->unison_shift;
#endif
#if UNISON_VOICES > 3
	osc->cur_sample[3] += osc->ds - 2*osc->unison_shift;
#endif
}
#define UNISON_SHIFT	((UNISON_VOICES+1) / 2)

inline s8 sample_osc(oscillator* osc) {
	s16 s1 = 0, s2 = 0;
	s8 i;
	u8 sample_pos;
	
	for(i=0; i<UNISON_VOICES; i++) {
		sample_pos = (osc->cur_sample[i] >> 10) & (WAVET_SIZE-1);

		s1 += wave_table1[ sample_pos ];
		s2 += wave_table2[ sample_pos ];

	}
	return (s1 >> UNISON_SHIFT) + ( (s2 >> UNISON_SHIFT)*osc->blend_val >> 8 );
}

inline s8 sample_drum(channel *chan) {
	if(chan->osc_c.ds < 32 || chan->vol == 0)
		return 0;

	chan->osc_c.ds--;

	return sample_osc(&chan->osc_c);
}
// End fast functions

// +==========================+
// |    Song Functionality    |
// +==========================+

inline u8 read_byte() {
#if SIMULATOR
	return song_data[song_pos++];
#else
	return read_song_byte(song_pos++);
#endif
}

inline u16 read_word() {
#if SIMULATOR
	u16 *p = (u16*)(song_data + song_pos);
	song_pos += 2;
	return *p;
#else
	u16 w = read_song_word(song_pos);
	song_pos += 2;
	return w;
#endif
}

// Mark a region with C_REPEAT=0 and C_REPEAT=x to repeat it x times.
inline void handle_repeat(u8 repeats_req) {
	if(!repeats_req) {
		repeat_pos = song_pos;
		repeats = 0;
	} else if(repeats_req > repeats) {
		song_pos = repeat_pos;
		repeats++;
	}		
}

inline void read_envelope(envelope *env) {
	env->init_state = read_byte();
	switch(env->init_state) {
		case ENV_ATTACK:
			env->initial = read_byte();
			env->attack = read_byte();
		case ENV_DECAY:
			env->max = read_byte();
			env->decay = read_byte();
		case ENV_HOLD:
			env->hold = read_byte();
	}
}

inline void read_events() {
	u8 header = read_byte();
	channel *chan = &g_channels[ header & 0xF ];

	switch(header >> 4) {
	case E_NOTE_ON:
		chan->vol = read_byte();
		//chan->pan_mask[0] = (chan->vol & 0x1) ? 0xFF : 0;
		//chan->pan_mask[1] = (chan->vol & 0x2) ? 0xFF : 0;
		chan->vol &= 0xFC;
		chan->vol <<= 1;
	case E_NOTE_ON_LAST_VOL:
		env_reset(&chan->env_vol);
		env_reset(&chan->env_pan);
		env_reset(&chan->osc_c.blend);
		//chan->note_duration = read_byte();
	case C_PITCH:
		//cli(); // no need to guard word reads
		chan->osc_c.ds = read_word();
		chan->osc_c.target_ds = chan->osc_c.ds;
		chan->osc_c.unison_shift = read_byte();
		//sei();
		PORTC |= LED_GREEN;
		break;
	case E_NOTE_OFF:
		chan->vol = 0;
		break;
	case C_MOD:
		chan0_mod = read_byte();
		break;
	case E_LED:
		PORTC &= ~(LED_GREEN | LED_RED);
		PORTC |= read_byte();
		break;
	case C_BLEND: // override blend
		chan->osc_c.blend.state = ENV_FORCED;
		chan->osc_c.blend.hold = read_byte();
		break;
	//case E_PROGRAM:
	//	chan->type = read_byte();
	//	break;
	case C_REPEAT:
		handle_repeat(read_byte());
		break;
	case E_VOL_ENV_SET:
		read_envelope(&chan->env_vol);
		break;
	case E_BL_ENV_SET:
		read_envelope(&chan->osc_c.blend);
		break;
	case E_END:
		song_pos = 0;
		repeats = 0;
		break;
	}
	time2event = read_byte();
}


inline void control() {
#if !SIMULATOR
	while(1) {
		if(control_signal) {
			
			control_signal = 0;
			u8 mask = LED_GREEN;
		//	if(clear_leds) {
		//		if(!clear_leds--)
					mask |= LED_RED;	
		//	}
			PORTC &= ~mask;
#endif
			
			if((jiffies & 0x1) == 0) {
				oscillator *o;
				if(demo) {
					while(!time2event) {
						read_events();
					}
					time2event--;
				}

				env_update_all();
				menu_update();

				if(chan0_mod) {
					// Channel 0 extras
					o = &g_channels[0].osc_c;
					// Modulation
					o->ds += (chan0_mod * sin_table[sin_index++] + 128) >> 8;
					sin_index &= 15;
				}
			}
			lcd_update();
			jiffies++;
#if !SIMULATOR
		}
		sleep_cpu();
	}
#endif
}

#if SIMULATOR
void synth_main()
#else
//#include <util/delay.h>
__attribute__((naked)) int main(void)
#endif
{
	u8 i;
	for(i = 0; i < NUM_CHANNELS; i++) {
		channel *chan = g_channels + i;
		chan->lfo_scale = NUM_CHANNELS - i;
		chan->env_vol.initial = 192;
		chan->env_vol.attack = 10;
		chan->env_vol.max = 255;
		chan->env_vol.decay = 2;
		chan->env_vol.hold = 192;
		chan->env_vol.init_state = ENV_ATTACK;
		chan->env_pan.initial = 16;
		chan->env_pan.attack = (i==0)?10:1;
		chan->env_pan.max = 205;
		chan->env_pan.decay = (i==0)?8:1;
		chan->env_pan.hold = 192;
		chan->env_pan.init_state = ENV_ATTACK;
		chan->osc_c.blend.initial = (i==0)?250:64;
		chan->osc_c.blend.attack = (i==0)?3:1;
		chan->osc_c.blend.max = 255;
		chan->osc_c.blend.decay = (i==0)?25:1;
		chan->osc_c.blend.init_state = ENV_ATTACK;
	}
	time2event = read_byte();
	time2control = trigger_val;

	lcd_init(version_string, 0);
#if !SIMULATOR
	// Configure ports
	DDRB = 0b00000110; // OC1AB are L/R PWM output
	PORTB = 0b00001001; // Activate pull-ups for the switches

	DDRC = 0b11011111; // Port C 2 and 3 are LED output, 5 is slider input (ADC)
	PORTC = 0b00;

	DDRD = 0xFE;	// PORTD is LCD out and USART in

	// Set clock to 8 MHz
	//CLKPR = 0x80; // Enable prescaler change
	//CLKPR = 0x80; // Scaling of 1 (20 MHz)

	// Set timer interrupt for sampling thread
	TCCR0A = 0b10; // Clear Timer on Compare (i.e. when TCNT0 == OCR0A)
	TCCR0B = 0b10; // Increment TCNT0 at CPU_SPEED/8 (= 2.5 MHz)
	OCR0A =  CPU_SPEED / 8 / SAMPLE_RATE;
	TIMSK0 = 0b10;

	// Set timer interrupt for control thread
	TCCR1A = 0b10100001; // non-inverting fast PWM on OC1AB
	TCCR1B = 0b00001001; // Increment TCNT2 at CPU_SPEED (20MHz / 256 = 78.125 KHz max sampling rate)
	//OCR1A = 0x80;
	//OCR1B = 0x80;

	// Set sleep mode
	SMCR = 0b0001; // Stop CPU clock (Ch. 10.9)
	PRR = 0b11010100; // Keep only timer 0, 1, ADC and USART on

	// Slider ADC
	ADMUX = 0b01100101; // ADC5, AVcc
	ADCSRA = 0b11000111; // ~16kHz sampling rate

	// MIDI IN (USART)
	UCSR0B = 0b10010000; // RX interrupt
	UCSR0C = 0b00000110; // 8bit words
	UBRR0L = 40; // 31.25kHz baud rate

	sei();
	control(); // Never returns
#endif
}

// Tone generation

#if SIMULATOR
void generate_sample(u8 *buf) {
#else
ISR(TIMER0_COMPA_vect) {
#endif
	u8 i, smp;
	s8 out[2] = {0,0};
repeat:
	for(i = 0; i < NUM_CHANNELS; i++) {
		update_osc(&g_channels[i].osc_c);
	}


	if(!time2control--) {
		time2control = trigger_val;
		control_signal = 1;
#if SIMULATOR
		control();
#else
		return; // skip a sample (1 in ~250) and yield to the control thread. Gives us +2KHz sampling rate.
#endif
	}

	for(i = 0; i < NUM_CHANNELS; i++) {
		channel *chan = g_channels + i;
		if(chan->inst_vol) {
			s8 smp = sample_osc(&chan->osc_c);
			out[0] += smp * chan->inst_vol[0] >> 8;
			out[1] += smp * chan->inst_vol[1] >> 8;
		}
	}
#if DRUM_CHAN>1
	smp = sample_drum(g_channels + DRUM_CHAN) * g_channels[DRUM_CHAN].vol >> 8;
	out[0] += smp & g_channels[DRUM_CHAN].pan_mask[0];
	out[1] += smp & g_channels[DRUM_CHAN].pan_mask[1];
#endif

#if SIMULATOR
	buf[0] = out[0] + 128;
	buf[1] = out[1] + 128;
#else
	OCR1A = out[0] + 128;
	OCR1B = out[1] + 128;
	if(TIFR0 & 0b10) { // Save interrupt overhead if timer already triggered Gives us +500Hz sampling rate
		TIFR0 = 0b10;
		LED_ON(200); // Turn on red LED to show overrun
		goto repeat;
	}
#endif
}

// MIDI IN

#include "tones.h"

enum midi_state {
	E_NOTE_ON_PITCH,
	E_NOTE_ON_VEL,
	E_NOTE_OFF_PITCH,
	E_NOTE_OFF_VEL,
	E_PITCHB_LSB,
	E_PITCHB_MSB,
	E_CONTOL_TYPE,
	E_CONTROL_CUTOFF,
	E_CONTROL_VOL_INIT,
	E_CONTROL_VOL_ATTACK,
	E_CONTROL_VOL_DECAY,
	E_CONTROL_VOL_HOLD,
	E_NO_STATE
};

u16 midi_portam_on_clock;
u8 midi_state;
u8 midi_chan;
u8 midi_chan_override; // for split key
u8 midi_cmd;
u8 midi_last_note[NUM_CHANNELS];
u8 midi_last_note2[NUM_CHANNELS]; // the note before last

#if !SIMULATOR
ISR(USART_RX_vect) {
	u8 midi_in = UDR0;
	u8 cmd;

	if(midi_in == MIDI_CLK) {
		return;
	}

	PORTC |= LED_GREEN;

	// cmd
	if(midi_in & 0x80) {
		cmd = midi_in>>4;
		midi_chan = midi_in & 0xF;
		switch(cmd) {
			case E_NOTE_ON:
				midi_cmd = cmd;
				midi_state = E_NOTE_ON_PITCH;
				break;
			case E_NOTE_OFF:
				midi_cmd = cmd;
				midi_state = E_NOTE_OFF_PITCH;
				break;
			case MIDI_PITCHBEND:
				midi_state = E_PITCHB_LSB;
				break;
			case MIDI_CONTROL:
				midi_state = E_CONTOL_TYPE;
				break;
			case MIDI_RESET:
				asm ("jmp 0x0000;");
			default:
				midi_state = E_NO_STATE;
				break;
		}
		return;
	}

	if(midi_chan >= NUM_CHANNELS)
		return;

	channel *chan = &g_channels[midi_chan_override];
	// data
	switch(midi_state) {
		case E_NOTE_ON_PITCH:
		{
			if(split_key > midi_in) {
				midi_in += 12;
				midi_chan_override = 1;
				chan = g_channels + 1;
			} else if(split_key > 0) {
				midi_chan_override = 0;
				chan = g_channels;
			}
			
			u8 note = midi_in - tones_offset - 12;
			chan->osc_c.target_ds = tones2offsets[note];
			chan->osc_c.unison_shift = tones2unison[note];

			if(!chan->vol) {
				env_reset(&chan->env_vol);
				env_reset(&chan->env_pan);
				env_reset(&chan->osc_c.blend);
				chan->osc_c.ds = chan->osc_c.target_ds;
				midi_last_note2[midi_chan_override] = 0;
			} else {
				// remember initial portamento note if we need to slide back
				midi_last_note2[midi_chan_override] = midi_last_note[midi_chan_override];
				// remember when portamento started so we can chop it up if overlap is not large enough
				midi_portam_on_clock = jiffies;
			}

			midi_last_note[midi_chan_override] = note;
			midi_state = E_NOTE_ON_VEL;
			if(menu_current_item() == M_MIDI_NOTE)
				lcd_print_value(midi_in);
			break;
		}
		case E_NOTE_OFF_PITCH:
		{
			if(split_key > midi_in) {
				midi_chan_override = 1;
				chan = g_channels + 1;
				midi_in += 12;
			} else {
				midi_chan_override = 0;
				chan = g_channels;
			}
			u8 note = midi_in - tones_offset - 12;

			if(note == midi_last_note2[midi_chan_override]) {
				midi_last_note2[midi_chan_override] = 0;
				if(jiffies - midi_portam_on_clock < 12) {
					env_reset(&chan->env_vol);
					env_reset(&chan->env_pan);
					env_reset(&chan->osc_c.blend);
					chan->osc_c.ds = chan->osc_c.target_ds;
				}
			} else
			if(note == midi_last_note[midi_chan_override]) { // current note is off
				if(midi_last_note2[midi_chan_override]) { // did we slide to here
					u8 to_note = midi_last_note2[midi_chan_override];
					chan->osc_c.target_ds = tones2offsets[to_note];
					chan->osc_c.unison_shift = tones2unison[to_note];
					midi_last_note[midi_chan_override] = to_note;
					midi_last_note2[midi_chan_override] = 0;
				}
				else
					chan->vol = 0;
			}
			midi_state = E_NOTE_OFF_VEL;
			break;
		}
		case E_NOTE_ON_VEL:
			if(midi_in > 0) {
				chan->vol = midi_in << 1;
			} else { // note OFF
				chan->vol = 0;
				midi_last_note[midi_chan_override] = 0;
			}
			midi_state = E_NOTE_ON_PITCH;
			break;
		case E_PITCHB_LSB:
			midi_state = E_PITCHB_MSB;
			break;
		case E_PITCHB_MSB:
			{
				u8 tone = midi_last_note[midi_chan];
				u16 v1 = tones2offsets[tone];
				u16 v2 = tones2offsets[tone + 2];
				s8 bend = midi_in - 64; // [-63; 63]
				chan->osc_c.ds = chan->osc_c.target_ds = v1 + ((s16)(v2 - v1) * bend >> 6);
				midi_state = E_NO_STATE;
			}
			break;
		case E_CONTOL_TYPE:
			//lcd_print_value(midi_in);
			switch(midi_in) {
				case 74: midi_state = E_CONTROL_CUTOFF; break;
				case 73: midi_state = E_CONTROL_VOL_INIT; break;
				case 19: midi_state = E_CONTROL_VOL_ATTACK; break;
				case 20: midi_state = E_CONTROL_VOL_DECAY; break;
				case 21: midi_state = E_CONTROL_VOL_HOLD; break;
				default: midi_state = E_NO_STATE; break;
			}
			break;
		case E_CONTROL_CUTOFF:
			chan->osc_c.blend.state = ENV_FORCED;
			chan->osc_c.blend.hold = (midi_in << 1) + 1;
			menu_set(M_B_HOLD);
			break;
		case E_CONTROL_VOL_INIT:
			chan->env_vol.initial = (midi_in << 1) + 1;
			menu_set(M_V_INITIAL);
			break;
		case E_CONTROL_VOL_ATTACK:
			chan->env_vol.attack = (midi_in << 1) + 1;
			menu_set(M_V_ATTACK);
			break;
		case E_CONTROL_VOL_DECAY:
			chan->env_vol.decay = (midi_in << 1) + 1;
			menu_set(M_V_DECAY);
			break;
		case E_CONTROL_VOL_HOLD:
			chan->env_vol.hold = (midi_in << 1) + 1;
			menu_set(M_V_HOLD);
			break;
	}

	return;
//	lcd_print_value(midi_in);
}
#endif
